#!/usr/bin/perl -w

# This program establishes a connection to lirc 
# and the Digital Audio Jukebox so you can control
# the music using an infra-red remote control.
# http://www.willcoxonline.com

use IO::Socket;
use IO::Select;
use strict;

my $select=IO::Select->new();

my $daj_server='192.168.1.13';
my $daj_port='7769';

my $lirc_server='192.168.1.13';
my $lirc_port='8765';

my $selected_zone='0'; 		# Set default zone
my $daj_server_handle;
my $lirc_server_handle;
my $buf;

if($daj_server_handle=IO::Socket::INET->new(Proto=>"tcp", PeerAddr=>$daj_server, PeerPort=>$daj_port))
	{
		print "Connected to Digital Audio Jukebox\n";
	}
	else
	{
		die "Cannot connect to Digial Audio Jukebox\n"; 
	};
if ($lirc_server_handle=IO::Socket::INET->new(Proto=>"tcp", PeerAddr=>$lirc_server, PeerPort=>$lirc_port)) 
	{
		print "Connected to lirc\n";
	}
	else
	{
		die "Cannot connect to lirc server\n"; 
	};

print "Waiting for commands.....\n";

$select->add($daj_server_handle);		# add the daj server socket handle to Select
$select->add($lirc_server_handle);		# add the lirc server socket handle to Select
my $socket_input="";
while (my @ready = $select->can_read,0.005) 	# Is there a handle waiting to be read?
{   
	foreach my $fh (@ready)			# Check each handle in turn
	{	
 		if (sysread ($fh,$buf,1) eq 0)  # Make sure socket hasn't disconnected
                {
			if ($fh eq $lirc_server_handle){die "Oops, lirc disconnected us!"};		
			if ($fh eq $daj_server_handle){die "Oops, daj disconnected us!"};		
                };
		if ($buf ne "\n")
                {
                	if ($buf ne "\r"){$socket_input=join "",$socket_input,$buf}; # Assemble the string
                }
                else
                {
                	if ($fh eq $lirc_server_handle)	# If the data was from lirc....
			{
				process_lirc_command(split(/ /,$socket_input)); # Process it.
			};
			$socket_input="";
                };
	}
};

sub process_lirc_command 
{
	my $ir_code=$_[0];
	my $ir_sequence_id=$_[1];
	my $ir_command=$_[2];
	my $ir_device=$_[3];
	
	print "Command: $ir_command Device: $ir_device\n";
	if ($ir_device eq 'daj' and $ir_sequence_id eq '00') # Only action the following commands once per button press
	{
		if ($ir_command eq 'play')
		{
			send_daj_command ("PLAYLIST:$selected_zone:PLAY");
		}
		elsif ($ir_command eq 'stop')
		{
			send_daj_command ("PLAYLIST:$selected_zone:STOP");
		}
		elsif ($ir_command eq 'skip')
		{
			send_daj_command ("PLAYLIST:$selected_zone:SKIP");
		}
		elsif ($ir_command eq '0')
		{
			$selected_zone='0';
		}
		elsif ($ir_command eq '1')
		{
			$selected_zone='1';
		}
		elsif ($ir_command eq '2')
		{
			$selected_zone='2';
		};
	};
	
	if ($ir_device eq 'daj') # These buttons can repeat if held down
	{
		if ($ir_command eq 'ff')
		{
			send_daj_command ("PLAYLIST:$selected_zone:SEEK+:2");
		}
		elsif ($ir_command eq 'rew')
		{
			send_daj_command ("PLAYLIST:$selected_zone:SEEK-:2");
		}
		elsif ($ir_command eq 'vol+')
		{
			send_daj_command ("PLAYLIST:$selected_zone:VOLUME+:5");
		}
		elsif ($ir_command eq 'vol-')
		{
			send_daj_command ("PLAYLIST:$selected_zone:VOLUME-:5");
		};
	};

};

sub send_daj_command
{
	my $command=$_[0];
	print $daj_server_handle "$command\n" or die "Connection to DAJ server broken!";
}
