#!/usr/bin/perl -w

#    Digital Audio Jukebox Server V1.0
#    Copyright (C) 2003  Ben Willcox (http://www.willcoxonline.com)
#
#    This program is free software; you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation; either version 2 of the License, or
#    any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program; if not, write to the Free Software
#    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

use strict;

use IO::Socket; # For remote connections
use IO::Select; # To check filehandles
use IPC::Open2; # Opens a bi-directional pipe (for mpg123)
use DBI;	# MySQL interface
use Audio::Mixer;
use POSIX ":sys_wait_h";
use FileHandle;


my $listen_port=7769;
my $db_user='daj_user';
my $db_password='secret_password';
my $logging='1'; # 1=on
my $music_dir = '/data/Music';
my $dbsync = './dbsync';
my $playback_engine = 'mpg321';
	

my @channels;
my $dbh=DBI->connect("DBI:mysql:daj",$db_user,$db_password) or die "Failed to connect to MySQL Server\n";
my $dbsync_pid=0;
my $listen = IO::Socket::INET->new (LocalPort=>$listen_port,
                                 Proto=>'tcp',
				 Reuse=>1,
                                 Listen=>5) || die "Cannot create socket - is server already running?";
my $select = new IO::Select() || die "Cannot select socket";
print "\n";
my $sync=0;

read_config();
my $no_channels=@channels-1;
foreach my $element (0..$no_channels)	# Open an instance of the playback engine for each channel 
{
	open_player($channels[$element]{channel_id},$channels[$element]{mpg123_write_fh},$channels[$element]{mpg123_read_fh},$element);
	$select->add($channels[$element]{mpg123_read_fh});
};

$select->add($listen);
#$SIG{'QUIT'}='close_program';
$SIG{'INT'}='close_program';
#$SIG{'HUP'}='close_program';
my $socket_input;
my $reader_input;

while (my @ready = $select->can_read) { # Is there a handle waiting to be read?
   foreach my $fh (@ready) 
     {	
      if ($fh eq $listen) # Connection received on listening socket 
	{
        	my $new = $listen->accept;
         	$select->add($new);
		new_client_connection($new);   
		print "Client connected, $new\n";	
	} 
	else # Check if data is waiting from MPG123 or clients
       	{
		my $foundit=0;
		foreach my $element (0..$no_channels)
		{
		if ($fh == $channels[$element]{mpg123_read_fh}) # Is it from MPG123?
			{
			player_read($channels[$element]{mpg123_read_fh});
			$foundit=1;
			};
		};
	 	 if ($foundit eq 0){socket_read($fh)}; # If no then it must be from a client
     	}
   }
};
#-------------------------------------------------------
#-------------------------------------------------------
sub open_player # <channel> <reader filehandle> <writer filehandle>
{
	my $channel=$_[0];
	my $output_device=$channels[$channel]{sound_device};		# Define audio device for this channel
	my $element=$_[3];						# Player number

	if ($channels[$element]{mpg123_pid}=open2($_[2], $_[1], "mpg321 -R -q -a $output_device --aggressive dummy"))
	{
		set_volume($_[0],0);				# Set volume to 0 
		print "MPG123 started pid: $channels[$element]{mpg123_pid}, output device:$output_device\n";
	};
};
#-------------------------------------------------------
sub close_player
{
	print "Closing players....\n";

	foreach my $element (0..$no_channels)
	{
	#	player_command('QUIT',$element);
		print "killing $channels[$element]{mpg123_pid}\n";
		kill 9, $channels[$element]{mpg123_pid};
	};
exit;
};
#-------------------------------------------------------
sub player_command # <command> <channel>
{
	my $command=$_[0];
	my $channel_id=$_[1];
	
	my $fh=$channels[$channel_id]{mpg123_write_fh};
	print $fh "$_[0]\n";
};
#-------------------------------------------------------
sub global_client_message
{
	foreach my $fh ($select->handles)
	{	
		my $hand_check=0;
		foreach my $element (0..$no_channels)
		{
			if ($fh eq $channels[$element]{mpg123_read_fh} ){$hand_check=1;};
		};
		if ($hand_check eq 0 && $fh ne \*STDIN && $fh ne $listen)
		{
			print $fh "$_[0]\n";
		};
	}
			logging("\c[\[32mSEND-GLOB: $_[0]\c[\[37m");
};
#-------------------------------------------------------
sub private_client_message # <message> <client filehandle>
{
	my $output_fh=$_[1];
	print $output_fh "$_[0]\n";
	logging ("\c[\[33mSEND-PRIV: $_[0]\c[\[37m");
};
#-------------------------------------------------------
sub player_read
{
	my $handle=$_[0];
	my $channel_id;
	foreach my $element (0..$no_channels)
	{
		if ($channels[$element]{mpg123_read_fh} eq $handle) 
		{
			$channel_id= $element;
		};	
	};

	my $player_total="";
	my $player_read="";
	while ($player_read ne "\n")
	{
		sysread($handle,$player_read,1);
		$player_total=join "",$player_total, $player_read;
	};
	my @player_data=split / /,$player_total;
	chomp @player_data;
	if ($player_data[0] eq '@P' && $player_data[1] eq '0' && $channels[$channel_id]{playermode} eq 'playing')
	{	
		delete_current_playlist_song($channel_id);
		get_next_playlist($channel_id);
	}
	elsif ($player_data[0] eq '@F')
	{
		$channels[$channel_id]{current_time_pos_frames} = $player_data[1];	
		if (int($channels[$channel_id]{current_time_pos_secs}) ne int($player_data[3]))
		{
			global_client_message("MSG_POS:$channel_id:".int($player_data[3]).":".int($player_data[4]).":".int(($player_data[1])+($player_data[2])));
		};
		$channels[$channel_id]{current_time_pos_secs} = $player_data[3];	
		if (($channels[$channel_id]{current_time_pos_secs} ne 0) && ($channels[$channel_id]{current_time_pos_frames} ne 0))
		{
			$channels[$channel_id]{frames_per_second}= int($channels[$channel_id]{current_time_pos_frames}/$channels[$channel_id]{current_time_pos_secs});
		};
	};
};
#-------------------------------------------------------
sub socket_read
{
	my $buf="";
	my $buf_total="";
	while ($buf ne "\n")  		# If its not EOL terminator...
	{
		if (sysread($_[0],$buf,1) eq 0)
		{
			$select->remove($_[0]);
			print "Client disconnected, $_[0]\n";	
			last;
		};
		if ($buf ne "\r"and $buf ne "\n") {$buf_total=join "",$buf_total, $buf};
	}
		if ($buf_total) {logging ("RECEIVED : $buf_total")};
	process_command ($buf_total,$_[0]);
};
#-------------------------------------------------------
sub process_command
{
	my ($arg1,$arg2,$arg3,$arg4);

	($arg1,$arg2,$arg3,$arg4)=split(/:/,uc($_[0]));	# Split the received command string at the : delimiter
	if ($arg1)
	{
		my $output_fh = $_[1]; # Filehandle to redirect output to.
		if ($arg1 eq 'GET')
		{
			if ($arg2 eq 'ARTISTS')
			{
				get_artists($output_fh);
			}
			elsif ($arg2 eq 'ALBUMS')
			{
				get_albums($arg3,$output_fh);
			}
			elsif ($arg2 eq 'TRACKS')
			{
				get_tracks($arg3,$arg4,$output_fh);
			}
			elsif ($arg2 eq 'PLAYLIST')
			{
				get_playlist($arg3,$output_fh);
			}
			elsif ($arg2 eq 'CHANNELS')
			{
				get_channels($output_fh);
			}
			elsif ($arg2 eq 'VOLUME')
			{
				private_client_message("MSG_VOLUME:$arg3:$channels[$arg3]{volume}",$output_fh);
				#private_client_message("MSG_OK",$output_fh);
			}
			else
			{
				private_client_message('ERR_INVALID_ARGUMENT',$output_fh);
			};
			
		}
		elsif ($arg1 eq 'PLAYLIST')
		{
			if ($arg3 eq 'ADD')
			{
				playlist_add_item($arg2,$arg4,$output_fh);
			}
			elsif ($arg3 eq 'COPY')
			{
				playlist_clear($arg2,$output_fh);
				playlist_copy($arg4,$arg2,$output_fh);
				playlist_play($arg2,$output_fh);
				player_jump($arg2,$channels[$arg4]{current_time_pos_frames});
			}
			elsif ($arg3 eq 'MOVE')
			{
				playlist_clear($arg2,$output_fh);
				playlist_copy($arg4,$arg2,$output_fh);
				playlist_clear($arg4,$output_fh);
				playlist_stop($arg4,$output_fh);
				playlist_play($arg2,$output_fh);
				player_jump($arg2,$channels[$arg4]{current_time_pos_frames});
			}
			elsif ($arg3 eq 'DELETE')
			{
				playlist_delete_item($arg2,$arg4,$output_fh);
			}		
			elsif ($arg3 eq 'CLEAR')
			{
				playlist_clear($arg2,$output_fh);
			}
			elsif ($arg3 eq 'PLAY')
			{
				playlist_play($arg2,$output_fh);
			}
			elsif ($arg3 eq 'PAUSE')
			{
				playlist_pause($arg2,$output_fh);
			}
			elsif ($arg3 eq 'STOP')
			{
				playlist_stop($arg2,$output_fh);
			}
			elsif ($arg3 eq 'SKIP')
			{
				playlist_skip($arg2,$output_fh);
			}	
			elsif ($arg3 eq 'SEEK-')
			{
				playlist_seek_back($arg2,$arg4,$output_fh);
			}	
			elsif ($arg3 eq 'SEEK+')
			{
				playlist_seek_forward($arg2,$arg4,$output_fh);
			}	
			elsif ($arg3 eq 'JUMP')
			{
				player_jump($arg2,$arg4,$output_fh);
			}
			elsif ($arg3 eq 'VOLUME')
			{
				set_volume($arg2,$arg4);
			}	
			elsif ($arg3 eq 'VOLUME+')
			{
				increase_volume($arg2,$arg4);
			}
			elsif ($arg3 eq 'VOLUME-')
			{	
				decrease_volume($arg2,$arg4);
			}
			else
			{
				private_client_message('ERR_INVALID_ARGUMENT',$output_fh);
			};
		}

		elsif ($arg1 eq 'DBSYNC')
		{
		if (waitpid($dbsync_pid,&WNOHANG))
		{	
			global_client_message('MSG_DBSYNC_STARTED');
			defined( $dbsync_pid = fork ) or die "Couldn't fork: $!";
				unless( $dbsync_pid ) {
			    	exec "nice -n 20 $dbsync $music_dir";
			        die "Couldn't exec $!";
				};
		}
		else
		{
			private_client_message('ERR_SYNC_IN_PROGRESS',$output_fh);
		};
				
		
		}	
		else
		{
			private_client_message('ERR_INVALID_COMMAND',$output_fh)
		};

	};
};
#-------------------------------------------------------
sub get_next_playlist
{
	my $channel_no=$_[0];
	my $sth=$dbh->prepare("SELECT song_id FROM daj_playlist WHERE playlist_id = $_[0] ORDER BY id");
	$sth->execute();
	my @song_id=$sth->fetchrow_array();
	if ($song_id[0])
	{
		my $sth=$dbh->prepare("SELECT filename FROM daj_files WHERE id = ?");
		$sth->execute($song_id[0]);
		my @row=$sth->fetchrow_array();
		warn "$row[0] $channel_no";
		player_command("LOAD $row[0]",$channel_no); # Start playing next song
		$channels[$channel_no]{playermode}='playing'; # Flag this channels playing mode as STOPPED
	}
	else
	{
		$channels[$channel_no]{playermode}='stopped'; # Flag this channels playing mode as STOPPED
	};
};
#-------------------------------------------------------
sub delete_current_playlist_song #<playlist no>
{
	my $sth=$dbh->prepare("SELECT id FROM daj_playlist WHERE playlist_id= $_[0] ORDER BY id");
	$sth->execute();
	my $song_id=$sth->fetchrow;
	$sth=$dbh->prepare("DELETE FROM daj_playlist WHERE id = ?");
	$sth->execute($song_id);
	global_client_message("MSG_DELETE_PLAYLIST:$_[0]:$song_id");
};

#-------------------------------------------------------
sub new_client_connection
{
};

#-------------------------------------------------------
#            Various program functions                 +
#-------------------------------------------------------
sub read_config
{
	my @values;
	
	open (CONFIG, "zones.conf") or die 'Cannot open config file zones.conf!';
	my $dummy=<CONFIG>;
	while (<CONFIG>)
	{
 		chomp;
		@values=split /:/,$_;
		push @channels,{channel_id=>$values[0],
		mpg123_write_fh=>FileHandle->new,
		mpg123_read_fh=>FileHandle->new,
		mpg123_pid=>'0',
		playermode=>'stopped',
		current_time_pos_secs=>'0',
		total_time_secs=>0,
		label=>$values[1],
		volume=>'0',
		sound_device=>$values[4],
		mixer=>$values[2],
		mix_device=>$values[3],
		current_time_pos_frames=>'0',
		frames_per_second=>'0'};
	};
	close (CONFIG);
};
#-------------------------------------------------------
sub close_program
{
	print "arggghhh\n";
	close_player();
	#close ($listen);
	#exit;
};
#-------------------------------------------------------
sub logging
{
	if ($logging eq '1')
	{
		print "$_[0]\n";
	};
};


#-------------------------------------------------------
#            Server command functions                  +
#-------------------------------------------------------
sub playlist_copy
{
	my $source_playlist_no=$_[0];
	my $dest_playlist_no=$_[1];
	my $filehandle=$_[2];
	
	my $sth=$dbh->prepare("SELECT song_id FROM daj_playlist WHERE playlist_id = ? ORDER BY id");
	$sth->execute($source_playlist_no);
	while (my $row=$sth->fetchrow_array)
	{
		playlist_add_item($dest_playlist_no,$row,$filehandle);
	}
};
#-------------------------------------------------------
sub playlist_clear # <playlist no>,<calling filehandle>
{
	my $playlist_no=$_[0];
	my $filehandle=$_[1];
	
	private_client_message("MSG_OK",$filehandle);
	my $sth=$dbh->prepare("SELECT COUNT(playlist_id) FROM daj_playlist WHERE playlist_id = ? ");
	$sth->execute($playlist_no);
	my $no_rows=$sth->fetchrow;
	for my $cnt (1 .. $no_rows)
	{
		delete_current_playlist_song($playlist_no);	
	};
};
#-------------------------------------------------------
sub playlist_delete_item # <playlist no>,<item no>,<calling filehandle>
{
	my $playlist_no=$_[0];
	my $item_no=$_[1];
	my $filehandle=$_[2];
	
	my $sth=$dbh->prepare("DELETE FROM daj_playlist WHERE id= ?");
	if ($sth->execute($item_no))
	{
		private_client_message("MSG_OK",$filehandle);
		global_client_message("MSG_DELETE_PLAYLIST:$playlist_no:$item_no");
	}
	else
	{
		private_client_message('MSG_ERROR',$filehandle);
	};
};
#-------------------------------------------------------
sub playlist_add_item # <playlist no>,<item no>,<calling filehandle>
{
	my $playlist_no=$_[0];
	my $item_no=$_[1];
	my $filehandle=$_[2];
	
	my $sth=$dbh->prepare("INSERT INTO daj_playlist VALUES ('',?,?)");
	if ($sth->execute($item_no,$playlist_no))
	{
		my $sth=$dbh->prepare("SELECT last_insert_id()");
		$sth->execute();
		my $last_insert_id=$sth->fetchrow;
		
		$sth=$dbh->prepare("SELECT artist, track FROM daj_files WHERE id=?");
		$sth->execute("$item_no");
		my @row=$sth->fetchrow_array;
		private_client_message("MSG_OK",$filehandle);
		global_client_message("MSG_ADD_PLAYLIST:$playlist_no");
		global_client_message("$item_no:$row[0]:$row[1]:$last_insert_id");		# song-id:Artist:Track:autoid
		global_client_message("MSG_OK");
	}
	else
	{
		private_client_message('MSG_ERROR',$filehandle);
	};
};
#-------------------------------------------------------
sub playlist_play # <playlist no>,<calling filehandle>
{
	my $playlist_no=$_[0];
	my $filehandle=$_[1];
	
	if ($channels[$playlist_no]{playermode} eq 'paused')
	{
		player_command('PAUSE',$playlist_no);
		$channels[$playlist_no]{playermode}='playing';
	}
	else
	{
		$channels[$playlist_no]{playermode}='playing';
		get_next_playlist($playlist_no);
		private_client_message("MSG_OK",$filehandle);
	};
};
#-------------------------------------------------------
sub playlist_stop # <playlist no>,<calling filehandle>
{
	my $playlist_no=$_[0];
	my $filehandle=$_[1];
	
	$channels[$playlist_no]{playermode}='stopped';
	player_command('STOP',$playlist_no);
	private_client_message("MSG_OK",$filehandle);
	global_client_message("MSG_POS:$playlist_no:0:0:0:0");
};
#-------------------------------------------------------
sub playlist_pause # <playlist no>,<calling filehandle>
{
	my $playlist_no=$_[0];
	my $filehandle=$_[1];
	
	if ($channels[$playlist_no]{playermode} eq 'paused')
	{
		$channels[$playlist_no]{playermode}='playing';
	}
	elsif ($channels[$playlist_no]{playermode} eq 'playing')
	{
		$channels[$playlist_no]{playermode}='paused';
	};
	player_command('PAUSE',$playlist_no);
	private_client_message("MSG_OK",$filehandle);
};
#-------------------------------------------------------
sub playlist_skip # <playlist no>,<calling filehandle>
{
	my $playlist_no=$_[0];
	my $filehandle=$_[1];
	
	delete_current_playlist_song($playlist_no);
	get_next_playlist($playlist_no);
	private_client_message("MSG_OK",$filehandle);
};
#-------------------------------------------------------
sub playlist_seek_back # <playlist no>,<value in seconds>,<calling filehandle>
{
	my $playlist_no=$_[0];
	my $value=$_[1];
	my $filehandle=$_[2];
	
	if (!$value) {$value=1};  # If no value specified assume 1 second
	my $seek_value=$channels[$playlist_no]{current_time_pos_frames}-($value*$channels[$playlist_no]{frames_per_second})."\n";
	player_command("JUMP $seek_value",$playlist_no);
	private_client_message("MSG_OK",$filehandle);
};
#-------------------------------------------------------
sub playlist_seek_forward # <playlist no>,<value in seconds>,<calling filehandle>
{
	my $playlist_no=$_[0];
	my $value=$_[1];
	my $filehandle=$_[2];
	
	if (!$value) {$value=1};  # If no value specified assume 1 second
	my $seek_value=$channels[$playlist_no]{current_time_pos_frames}+($value*$channels[$playlist_no]{frames_per_second})."\n";
	player_command("JUMP $seek_value",$playlist_no);
	private_client_message("MSG_OK",$filehandle);
};
#-------------------------------------------------------
sub player_jump # <playlist no>,<value>,<calling filehandle>
{
	my $playlist_no=$_[0];
	my $value=$_[1];
	my $filehandle=$_[2];
	
	player_command("JUMP $value",$playlist_no);
	if ($filehandle)
	{
		private_client_message("MSG_OK",$filehandle);
	};
};	
#-------------------------------------------------------
sub get_artists # <calling filehandle>
{
	my $filehandle=$_[0];
	
	private_client_message("MSG_ARTISTS",$filehandle);
	my $sth=$dbh->prepare("SELECT DISTINCT artist FROM daj_files ORDER BY artist");
	$sth->execute();
	while (my @tmp=$sth->fetchrow_array) 
	{
		private_client_message("$tmp[0]",$filehandle);
	}; 
	private_client_message("MSG_OK",$filehandle);
};
#-------------------------------------------------------
sub get_albums # <artist>,<calling filehandle>
{
	my $artist=$_[0];
	my $filehandle=$_[1];
	
	if (!$artist)
	{
		private_client_message('ERR_NULL_ARGUMENT',$filehandle);
	}
	else
	{
		private_client_message('MSG_ALBUMS',$filehandle);
		my $sth=$dbh->prepare("SELECT DISTINCT album FROM daj_files WHERE artist= ? ORDER BY album");
		$sth->execute($artist);	
		while (my @tmp=$sth->fetchrow_array) 
		{
			private_client_message("$tmp[0]",$filehandle);
		}; 
		private_client_message('MSG_OK',$filehandle);
	}
};
#-------------------------------------------------------
sub get_tracks # <album>,<artist>,<calling filehandle>
{
	my $artist=$_[0];
	my $album=$_[1];
	my $filehandle=$_[2];

	if (!$artist||!$album)
	{
		private_client_message('ERR_NULL_ARGUMENT',$filehandle);
	}
	else
	{
		private_client_message('MSG_TRACKS',$filehandle);
		my $sth=$dbh->prepare("SELECT id, track, songnb FROM daj_files WHERE artist= ? AND album= ? ORDER BY songnb, track");
		$sth->execute($artist,$album);	
		while (my @tmp=$sth->fetchrow_array) 
		{
			if ($tmp[2])
			{
				$tmp[2] = $tmp[2]." - ";
			}
			else
			{
				$tmp[2] ="";
			};
			private_client_message("$tmp[0]:$tmp[2]$tmp[1]",$filehandle);
		}; 
		private_client_message('MSG_OK',$filehandle);
	};
};
#-------------------------------------------------------
sub get_playlist # <playlist no>,<calling filehandle>
{
	my $playlist_no=$_[0];
	my $filehandle=$_[1];
	
	my $sth=$dbh->prepare("SELECT song_id, artist, track, daj_playlist.id FROM daj_playlist INNER JOIN daj_files ON daj_playlist.song_id=daj_files.id WHERE playlist_id = ? ORDER BY daj_playlist.id");
	if ($sth->execute($playlist_no))
	{
		private_client_message("MSG_PLAYLIST:$playlist_no",$filehandle);
		while (my @row=$sth->fetchrow_array())
		{
			private_client_message("$row[0]:$row[1]:$row[2]:$row[3]",$filehandle);
		};
		private_client_message('MSG_OK',$filehandle);
	}
	else
	{
		private_client_message('MSG_ERROR',$filehandle);
	};	
};
#-------------------------------------------------------
sub get_channels # <calling filehandle>
{
	my $filehandle=$_[0];
	
	private_client_message("MSG_CHANNELS",$filehandle);
	foreach my $element (0..$no_channels)
	{
		private_client_message("$channels[$element]{channel_id}:$channels[$element]{label}",$filehandle);
	};
	private_client_message('MSG_OK',$filehandle);
};	
#-------------------------------------------------------
sub set_volume # <playlist no>, <volume>
{
	my $playlist_no=$_[0];
	my $volume=$_[1];
	
	$channels[$playlist_no]{volume}=$volume; 		# Store current volume for this channel 
	Audio::Mixer::set_mixer_dev($channels[$playlist_no]{mixer});	# Set audio control to correct mixer channel 
	Audio::Mixer::set_cval($channels[$playlist_no]{mix_device},$volume);			# Set volume
	global_client_message("MSG_VOLUME:$playlist_no:$channels[$playlist_no]{volume}");
};
#-------------------------------------------------------
sub increase_volume # <plaulist no>, <volume>
{
	my $playlist_no=$_[0];
	my $value=$_[1];
	
	my $new_volume=$channels[$playlist_no]{volume}+$value;
	set_volume($playlist_no,$new_volume);
};
#-------------------------------------------------------
sub decrease_volume # <plaulist no>, <volume>
{
	my $playlist_no=$_[0];
	my $value=$_[1];
	
	my $new_volume=$channels[$playlist_no]{volume}-$value;
	set_volume($playlist_no,$new_volume);
};
#-------------------------------------------------------
